//*************************************************************************************************************
// 
// Тагильцев Геннадий aka Marks
// версия v5
//
// изменения v2:
// - добавлена функция boolean Engine_Proc (int, int)
// параметры 
// - требуемое положение газа
// - условная задержка (в циклах программы)
// возвращает
// true - при достижении требуемого значения
// изменения v3
// - убрал процедуру таймера, вся обработка в одном цикле.
// изменения v4
// - расчет углов только на основании данных акселерометра
// изменения v5
// - добавляем программатор черех ГФК с помощью Arduino Uno + LCD Key shield
// - попробую другую библиотеку для MPU6050
// изменения v6
// - пробую добавить гувернер
// - гироскоп пока отключил
//
//**************************************************************************************************************
//#define _landingGear      // если ретрактов нет, ставим коммент на эту стоку строку
//#define _debugOn          // если отладка нужна убрать коммент
//#define _gyroOn           // если гироскопа нет, ставим коммент на эту стоку строку
//#define CLEAR_EEPROM

#include "_blinker.c"
#include "_button.c"

#include <Servo.h>
#include <EEPROM.h>


#ifdef _debugOn
#include <SoftwareSerial.h>
#endif

#ifdef _gyroOn
#include <Wire.h>
#include "Kalman.h"
#endif

#define pinButton 11                // порт сервисной кнопки
#define pinMotorESC 9               // порт к которому подключен ESC

#define Motor_RPM 4000              // номинальные обороты мотора в полете об/мин
#define Motor_ESC_Min_THR 10        // минимальны уровень газа, что бы гувернер не остановил мотор, в процентах
#define Motor_ESC_Start_THR 70      // уровень газа при старте модели, в процентах

#define Start 5                     // время отведенное на добежать до ручки, в течении этого времени процедура старта может быть отменена, в секундах  
#define Takeoff 5                  // время отведенное на разгон и взлет, в течении этого времени процедура старта может быть отменена, в секундах  
#define GearUp 5                   // время после начала разгона, начало уборки шасси -  5 секунд 
#define InFly 20                    // время полета в секундах с момента завершения процедуры старта - сейчас стоит 20 секунд

#define delay_g 100                 // коэффициент торможения ))
#define Engine_POL 7                // количество полюсов у мотора разделенное на два


#ifdef _landingGear
#define Port_Left_Wheel 6           // порт к которому подключена серва ретракта левого шасси
#define Port_Right_Wheel 7          // порт к которому подключена серва ретракта правого шасси
#define Left_Wheel_Down 2000
#define Left_Wheel_Up 1000
#define Right_Wheel_Down 1000
#define Right_Wheel_Up 2000
#define Up 1   
#define Down 0                      
#endif

  Servo Motor_ESC;                    // объект Servo для регулятора мотора
//  Button button (pinButton);      // объект кнопка управления             

#ifdef _landingGear
  Servo Left_Wheel;                   // объект Servo для ретракта левого шасси
  Servo Right_Wheel;                  // объект Servo для ретракта правого шасси
#endif

#ifdef _debugOn
  SoftwareSerial mySerial(5, 6);    // RX, TX>
#endif


//**************************************************************************************
// переменные определяющие ветвление программы
// проще говоря, состояние модели
//**************************************************************************************
  boolean _abort = false;                // процедура старта может быть прервана повторным нажатием кнопки 

#ifdef _gyroOn
//************************************************************************************
// блок объявления переменных для процедуры кальман
//************************************************************************************
Kalman kalmanX;
Kalman kalmanY;

uint8_t IMUAddress = 0x68;

/* IMU Data */
int16_t accX;
int16_t accY;
int16_t accZ;
int16_t tempRaw;
int16_t gyroX;
int16_t gyroY;
int16_t gyroZ;

double accXangle; // Angle calculate using the accelerometer
double accYangle;
double temp;
double gyroXangle = 180; // Angle calculate using the gyro
double gyroYangle = 180;
double compAngleX = 180; // Calculate the angle using a Kalman filter
double compAngleY = 180;
double kalAngleX; // Calculate the angle using a Kalman filter
double kalAngleY;
#endif

uint32_t timer;
uint32_t timeOut = 0;
uint8_t uartBufer = 0;
boolean connectFlag = false;
boolean flagTransfer = false;

//************************************************************************************
//  массив настроек программы (скоро уберу)
//************************************************************************************
  uint16_t stack[14] 
  {
  Motor_RPM,                  // обороты мотора в полете
  Motor_ESC_Min_THR,          // минимальный уровень газа, в процентах от максимального
  Motor_ESC_Start_THR,        // уровень газа на старте в процентах от максимального
  Engine_POL,                 // полюсов у мотора
  0,
  Start,                      // время на "убежать к ручке" в секундах    
  Takeoff,                    // время разгона и  взлета                  
  GearUp,                     // момент уборки шасси с момента начала разгона
  InFly,                      // общее время полета с момента старта двигателя минуты секунды
  0,
  0,
  0,
  0,
  0
  };

uint8_t* adrStack = (uint8_t*) stack;
uint8_t chkSumm = 0;
uint8_t i=0;
  
//************************************************************************************
// временны'е переменные 
//************************************************************************************
  unsigned long timeCurrent;         // текущее время, мало ли где то пригодится
  unsigned long timeStart;           // время старта по таймеру (т.е. момент нажатия на кнопку)
  unsigned long timeTakeOff;         // время начала разгона
  unsigned long timeGearUP;          // время начала уборки шасси (т.е. с момента запуска двигателя)
  unsigned long timeOfFlight;        // время завершения полета с момента запуска двигателя
//*************************************************************************************

//*************************************************************************************
// эти переменные нужны что бы не задаваться вопросом "А как у нас работают сервы"
//*************************************************************************************
  int Step_Left;
  int Step_Right;
//*************************************************************************************
const int ledPin = LED_BUILTIN;  // the number of the LED pin


//**********************************************************************************************************************
// переменные в интересах гувернера
//**********************************************************************************************************************
volatile uint32_t isr_time = 0;                   // прерывание пришло - время
volatile uint32_t isr_prev = 0;                   // предыдущее прерывание - время
volatile uint16_t isr_period = 0;                 // время на один оборот в микросекундах

volatile uint32_t isr_period_buffer[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; // делаем буфер на 16, так проще вычислять среднее
volatile uint32_t isr_period_buffer_summ = 0;     // сумма значений в бувере 
volatile uint32_t isr_period_buffer_average = 0;  // среднее значение в буфере
volatile uint16_t isr_period_delta = 0;           // максимальная ошибка 
volatile int16_t  isr_period_change = 0;          // отклоенеи от среднего в буфере
volatile uint32_t isr_period_min = 0;             // среднее значение в буфере
volatile uint32_t isr_period_max = 0;             // среднее значение в буфере
volatile uint8_t  isr_buffer_count = 0;           // счетчик буфера прерываний
volatile uint16_t isr_delta = 0;                  // допустимое отклонение периода одного оборота в микросекундах

uint32_t time_of_circle_buffer[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; // делаем буфер на 16, так проще вычислять среднее
uint32_t time_of_circle = 0;  // считаем время -количество полюсов- импульсов
uint8_t count_of_circle = 0;

volatile boolean _isrFlag = false;     // прерывание обработано

uint16_t last_PWM = 0;

uint32_t time_of_one_turn = 0;   // период одного оборота мотора
uint32_t time_of_one_turn_min = 0;   
uint32_t time_of_one_turn_max = 0;   
uint16_t delta_time_of_one_turn = 0;
uint32_t last_time = 0;

uint16_t minTHR;
uint16_t startTHR;

//*************************************************************************************
// инициализация таймера
//*************************************************************************************
void setup() 
{
  _setupButton(pinButton);
  _setupLed(LED_BUILTIN);
#ifdef _debugOn
  mySerial.begin(57600);
#endif
  Serial.begin(57600);

  engine_init();                          // инициализируем порт мотора
  readEEPROM();                           // читаем данные из EEPROM
  progMode();                             // идем в режим программирования

#ifdef _gyroOn
  Wire.begin();                           // инициализация i2c  
#endif
  minTHR = 1000 + stack[1] * 10;
  startTHR = 1000 + stack[2] * 10;
  Serial.println(minTHR);
  Serial.println(startTHR);
#ifdef _debugOn
  mySerial.println(minTHR);
  mySerial.println(startTHR);
#endif
  last_PWM = startTHR;                    // stack[2];                    // очень маленькие обороты, что бы заработал гувернер
  attachInterrupt(0, isr, RISING);        // прерывание на D2 
  init_gov(stack[0]);

#ifdef _landingGear
//**********************************************************************************
//* определение шага выпуска/уборки шасси
//**********************************************************************************
  int i = Left_Wheel_Down-Left_Wheel_Up;
  if (i>0) {
    Step_Left = -1;
  }
  else {
    Step_Left = 1;
  }
  i = Right_Wheel_Down-Right_Wheel_Up;
  if (i>0) {
    Step_Right = -1;
  }
  else {
    Step_Right = 1;
  }

    Left_Wheel.attach(Port_Left_Wheel);      
    delay(200);
    Right_Wheel.attach(Port_Right_Wheel);    
    delay(200);
    while (!Gear_Proc(Down));             // выпустили шасси
#endif

//**********************************************************************************
// предварительная настройка завершена, ждем нажатия кнопки старта
//*********************************************************************************

  while (!_button(pinButton, 200)) _blinker(LED_BUILTIN, 100, 1000);            // digitalRead(pinButton));
  while (_button(pinButton, 200)) _blinker(LED_BUILTIN, 1000, 100);            // digitalRead(pinButton));
  last_time = millis();
  timeStart = last_time + stack[5]*1000;
  timeTakeOff = timeStart + stack[6]*1000;    
  timeGearUP = timeTakeOff + stack[7]*1000;
  timeOfFlight = timeTakeOff + stack[8]*1000;

  
//*********************************************************************************

#ifdef _gyroOn
//**********************************************************************************
// блок setup для kalman
//**********************************************************************************
  i2cWrite(0x6B,0x00); // Disable sleep mode  
  if(i2cRead(0x75,1)[0] != 0x68)                  // Read "WHO_AM_I" register
  {
    Serial.print(F("MPU-6050 with address 0x"));
    Serial.print(IMUAddress,HEX);
    Serial.println(F(" is not connected"));
    while(1)
    {
      digitalWrite(13, LOW);
      delay(500);
      digitalWrite(13, HIGH);
      delay(1500);
    }
  }      
  kalmanX.setAngle(180); // Set starting angle
  kalmanY.setAngle(180);
  timer = micros();
#endif
}

// **********************************************************************************
// основной цикл программы
// **********************************************************************************
void loop() 
{
  while (millis() < timeStart && !_abort) 
  {
    _blinker(LED_BUILTIN, 200,1000);
    if (_button(pinButton, 200)) _abort = true;
  }

  while (millis() < timeTakeOff && !_abort) 
  {
    while (!Engine_Proc(last_PWM,1000));
    _blinker(LED_BUILTIN, 500, 200);
    if (_button(pinButton, 200)) _abort = true;
  }
  if (_abort)
  {
    Motor_ESC.writeMicroseconds(1000);
    while(1) _blinker(LED_BUILTIN, 1000,1000);
  }

#ifdef _landingGear  
  while (millis() < timeGearUP) Gear_Proc(Up);
#endif

  while (millis() < timeOfFlight)
  {
    Gouverner();
  }      

  Motor_ESC.writeMicroseconds(1000);
  while(1) _blinker(LED_BUILTIN, 200, 2000);

/* 
  {
  int currentTHR = Motor_ESC.readMicroseconds();

  Angel();
    if (accXangle>190 &&  accXangle<350) 
    {
     if (currentTHR != stack[3]) 
     {
       Motor_ESC.writeMicroseconds(stack[3]);
     }
   }
   else if (accXangle<170 &&  accXangle>10) 
   {
    if (currentTHR != stack[1]) 
    {
      Motor_ESC.writeMicroseconds(stack[1]);
    }
  }
  else 
  {
    Motor_ESC.writeMicroseconds(stack[2]);
  }
*/
/*
  }  
  while(millis() < Landing_Time) {
    if (Engine_Proc(stack[1],10)) Gear_Proc(Down);
  }
  
  while (millis() >= Landing_Time) Engine_Proc(stack[0],10) ;
*/   
}


#ifdef _landingGear
//*************************************************************************************
// процедура уборки/выпуска шасси 
// результат 
// - TRUE - завершено
// - FALSE - не завершена
// параметр - int - 
// - 0 - выпускам
// - 1 - убираем        
//*************************************************************************************
boolean Gear_Proc(int l) {
  static int i = delay_g;
  int j = Left_Wheel.readMicroseconds();
  int k = Right_Wheel.readMicroseconds();
if (l == 0) {          
// шасси выпускаем
  if (j == Left_Wheel_Down && k == Right_Wheel_Down) return true;

  if (i==0) {
    if (j > Left_Wheel_Down) j--;
    else j++;
    Left_Wheel.writeMicroseconds(j);
  }
    
  if (i==0) {
    if (k > Right_Wheel_Down) k--;
    else k++;
    Right_Wheel.writeMicroseconds(k);
  }
  
  if (!i--) i=delay_g;
}
else {
  // шасси выпускаем
  if (j == Left_Wheel_Up && k == Right_Wheel_Up) return true;

  if (i==0) {
    if (j > Left_Wheel_Up) j--;
    else j++;
    Left_Wheel.writeMicroseconds(j);
  }
    
  if (i==0) {
    if (k > Right_Wheel_Up) k--;
    else k++;
    Right_Wheel.writeMicroseconds(k);
 }
  if (!i--) i=delay_g;
}
  return false;
}
#endif



//*************************************************************************************
// Процедура работы с регуляторм
// возвращает true когда результат достигнут
// первый параметр - int - значение до которого надо изменить газ
// второй параметр - int - задержка
//*************************************************************************************
boolean Engine_Proc(int j1, int i1) {
  static int i = 0;
  int j =  Motor_ESC.readMicroseconds();

  if (j==j1) return true;
  
  if (i == i1 && j<j1) {
    j++;   
    Motor_ESC.writeMicroseconds(j);
  }
  if (i == i1 && j>j1) {
    j--;   
    Motor_ESC.writeMicroseconds(j);
  }
  if (!i--) i=i1;
//  if (digitalRead(pinButton) == LOW) _abort = true ;  
  return false;  
}



#ifdef _gyroOn
//***************************************************************************************
// кальман для гироскопа
// считает углы ))
//***************************************************************************************
void Angel() {
  uint8_t* data = i2cRead(0x3B,14);  
  accX = ((data[0] << 8) | data[1]);
  accY = ((data[2] << 8) | data[3]);
  accZ = ((data[4] << 8) | data[5]);  
  /* Calculate the angls based on the different sensors and algorithm */
  accYangle = (atan2(accX,accZ)+PI)*RAD_TO_DEG;
  accXangle = (atan2(accY,accZ)+PI)*RAD_TO_DEG;    
  delay (1);
}
// запись i2c
void i2cWrite(uint8_t registerAddress, uint8_t data){
  Wire.beginTransmission(IMUAddress);
  Wire.write(registerAddress);
  Wire.write(data);
  Wire.endTransmission(); // Send stop
}
// чтение i2c
uint8_t* i2cRead(uint8_t registerAddress, uint8_t nbytes) {
  uint8_t data[nbytes];  
  Wire.beginTransmission(IMUAddress);
  Wire.write(registerAddress);
  Wire.endTransmission(false); // Don't release the bus
  Wire.requestFrom(IMUAddress, nbytes); // Send a repeated start and then release the bus after reading
  for(uint8_t i = 0; i < nbytes; i++)
    data[i] = Wire.read();
  return data;
} 
#endif

//*************************************************************************************
// чтение EEPROM
//*************************************************************************************
void readEEPROM()
{
  chkSumm = 0;
  int i = 0;
#ifdef _debugOn
    mySerial.println("Read EEPROM");
#endif
#ifdef CLEAR_EEPROM
  EEPROM.write(0, 0xFF); //EEPROM[0] = 0xFF;
  delay(100);
#endif
  if (EEPROM[0] == 0xFF)
  {
    for (i=0; i <= 27; i++) EEPROM.write(i, adrStack[i]);  // EEPROM[i] = adrStack[i];
  }

  for (i=0; i <= 27; i++)
  {
    adrStack[i] = EEPROM.read(i);
    chkSumm = chkSumm + adrStack[i];
#ifdef _debugOn
    mySerial.print(adrStack[i],HEX);
    mySerial.print("<->");
    mySerial.println(stack[i/2],HEX);
#endif
  } 
}

//***************************************************************************
// прием данных от программатора
//***************************************************************************
void resiveSetting() 
{
  chkSumm = 0;
  timer = millis();
  i=0;
  uartBufer = 0;
  while (1)
  {
  //  digitalWrite(ledPin, HIGH);                       // зажгли светодиод
    if (millis() - timer > 500) 
    {
      return;
    }
    if (Serial.available() > 0) 
    {
      uartBufer = Serial.read();
#ifdef _debugOn
      mySerial.print("<-");
      mySerial.println(uartBufer,HEX);
#endif
      adrStack[i] = uartBufer;
      chkSumm = chkSumm + uartBufer;
      i++;
      if (i == 28) break; //  return;
    } 
  }
  while (1)
  {
    if (millis() - timer > 500) 
    {
//      digitalWrite(ledPin, HIGH);                       // зажгли светодиод
      return;
    }
    if (Serial.available() > 0)                         // ждем контрольную сумму
    {
      uartBufer = Serial.read();
#ifdef _debugOn
      mySerial.print("<-");
      mySerial.println(uartBufer,HEX);
#endif
      if (uartBufer == chkSumm) 
      {
//        digitalWrite(ledPin, LOW);                      // погасили светодиод
        for (i=0; i<=27; i++) EEPROM.write(i, adrStack[i]); //EEPROM[i] = adrStack[i];
        Serial.write (0x99);
#ifdef _debugOn
        mySerial.print("->");
        mySerial.println(0x99,HEX);
#endif
        return;
      }
      else
      {
        Serial.write (0xAA);
#ifdef _debugOn
        mySerial.print("->");
        mySerial.println(0xAA,HEX);
#endif
        return;
      }
    } 
  }
}

//***************************************************************************
// передача данных в программатор
//***************************************************************************

void sendSetting()
{
  chkSumm = 0;
  for (int i = 0; i <= 27; i++)
  {
    Serial.write(adrStack[i]);                  // отправили байт
#ifdef _debugOn
    mySerial.print("->");
    mySerial.println(adrStack[i],HEX);
#endif
    chkSumm = chkSumm + adrStack[i];            // считаем контрольную сумму
  }
  Serial.write(chkSumm);
#ifdef _debugOn
    mySerial.print("->");
    mySerial.println(chkSumm,HEX);
#endif
}

//***********************************************************************************
// режим программирования
//***********************************************************************************
void progMode () 
{
  i = 0;  
  timeOut = millis();
  connectFlag = false;
  uartBufer = 0;
  while (Serial.available()) Serial.read();
  #ifdef _debugOn
      mySerial.println("Clean");
#endif

  while (1)
  {
//    _blinker(LED_BUILTIN,500,1000);
    digitalWrite(ledPin, HIGH);                       // подтянули кнопку к +5V
    if (millis()-timeOut > 2000 && !connectFlag)        // блок таймаута при включении
    {
      digitalWrite(ledPin, LOW);                       // подтянули кнопку к +5V
      return;                                           // случка программатора и таймера не состоялась в течении 3  секунд после включения таймера
    }
    if (Serial.available()>0)                           // ждем данные из UART
    {
      uartBufer = Serial.read();
#ifdef _debugOn
      mySerial.print("<-");
      mySerial.println(uartBufer,HEX);
#endif
      switch (uartBufer)
      {
      case 0x55:                                        // запрос соединения
        Serial.write (0x55);
        Serial.flush();
        connectFlag = true;
#ifdef _debugOn
        mySerial.print("->");
        mySerial.println(0x55,HEX);
#endif
        break;
      case 0x5A:                                        // запрос настроек
        sendSetting();                                  // отправляем настройки
        break;
      case 0xA5:                                        // передача настроек от программатора
        resiveSetting();                                // принимаем настройки
      break;
      default:
      break;
      }
    }
  }
}

//***************************************************************************************************************
// обработчик импульсов
//***************************************************************************************************************
void isr() 
{
  isr_time = micros();
  _isrFlag = true;
}

//***************************************************************************************************************
// регулировка оборотов
//***************************************************************************************************************
void Gouverner() 
{
// пришло прерывание от мотора
  if (_isrFlag) 
  {
    _isrFlag = false;
    isr_period = isr_time - isr_prev;
    isr_period_buffer [isr_buffer_count] =   isr_period ;
    isr_prev = isr_time;
    isr_buffer_count++;
    if (isr_buffer_count > 15) isr_buffer_count = 0;

// кольцевой буфер еще не заполнен
    if (isr_period_buffer[15] == 0) return;
    isr_period_buffer_summ = 0;

    for (int i = 0; i<16; i++) isr_period_buffer_summ = isr_period_buffer_summ + isr_period_buffer[i];

    isr_period_buffer_average = isr_period_buffer_summ >> 4;
    isr_period_delta = isr_period_buffer_average >> 3;
    isr_period_max = isr_period_buffer_average + isr_period_delta;
    isr_period_min = isr_period_buffer_average - isr_period_delta;

    if ((isr_period > isr_period_max) || (isr_period < isr_period_min))   return;

//**********************************************************************************************************
// считаем текущие обороты
//**********************************************************************************************************
    time_of_circle_buffer[count_of_circle] = isr_period;
    count_of_circle++;
    if (count_of_circle == stack[3]) 
    {
      time_of_circle = 0;
      for (int i = 0; i < stack[3]; i++) time_of_circle = time_of_circle + time_of_circle_buffer[i]; 
      if (time_of_circle > time_of_one_turn_max) 
      { 
        last_PWM = last_PWM + 3;
        if (last_PWM > 2000) last_PWM = 2000;
//        if (last_PWM > Motor_ESC_Max_THR) last_PWM = Motor_ESC_Max_THR;
      }
      else if (time_of_circle < time_of_one_turn_min)
      {
        last_PWM = last_PWM - 3;
        if (last_PWM < minTHR) last_PWM = minTHR;
//        if (last_PWM < Motor_ESC_Landing_THR) last_PWM = Motor_ESC_Landing_THR;
      } 
      Motor_ESC.writeMicroseconds(last_PWM);
      count_of_circle = 0;
    }    
    if (time_of_circle_buffer[stack[3]-1] == 0) return;
  }
}

//***************************************************************************************************************
// модуль инициализации гувернера
//***************************************************************************************************************
void init_gov(uint16_t rpm)
{
  time_of_one_turn = 60000000 / rpm;
  delta_time_of_one_turn = time_of_one_turn >> 6;
  time_of_one_turn_min = time_of_one_turn - delta_time_of_one_turn;
  time_of_one_turn_max = time_of_one_turn + delta_time_of_one_turn;
}

//***************************************************************************************************************
// инициализация мотора
//***************************************************************************************************************
void engine_init() {
  Motor_ESC.attach(pinMotorESC, 1000, 2000);  // привязали регулятор к каналу
  delay(1000);
  if (_button(pinButton, 200))
  {
    Motor_ESC.writeMicroseconds(2000);
    delay (2500);                                                   // ждем 2 секунды
  }
  Motor_ESC.writeMicroseconds(1000);
  delay(1500);
  while (_button(pinButton, 200)) _blinker (LED_BUILTIN, 1000, 1000);
}

